package cc.toph.simplycompat.config;
/*
 * Copyright (c) 2021 magistermaks
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

import cc.toph.simplycompat.SimplyCompat;
import cc.toph.simplycompat.SimplyCompatExpectPlatform;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Map;
import java.util.Scanner;

import static cc.toph.simplycompat.SimplyCompat.LOGGER;

public class SimpleConfig {

  public static final String FILE_EXTENSION = ".config";
  private final HashMap<String, String> config = new HashMap<>();
  private final ConfigRequest request;
  private boolean broken = false;

  private SimpleConfig(ConfigRequest request) {
    this.request = request;
    String identifier = "Config '" + request.filename + "'";

    if (!request.file.exists()) {
      LOGGER.info("{} is missing, generating default one...", identifier);

      try {
        createConfig();
      } catch (IOException e) {
        LOGGER.error("{} failed to generate!", identifier);
        LOGGER.trace(e);
        broken = true;
      }
    }

    if (!broken) {
      try {
        loadConfig();
      } catch (Exception e) {
        LOGGER.error("{} failed to load!", identifier);
        LOGGER.trace(e);
        broken = true;
      }
    }

  }

  /**
   * Creates new config request object, ideally `namespace`
   * should be the name of the mod id of the requesting mod
   *
   * @param filename - name of the config file
   * @return new config request object
   */
  public static ConfigRequest of(String filename) {
    // File
    // Path path = SimplyCompatExpectPlatform.getConfigDirectory().resolve(filename + FILE_EXTENSION);
    // Folder
    Path path = SimplyCompatExpectPlatform.getConfigDirectory().resolve(String.format("%s/%s%s", SimplyCompat.MOD_ID, filename, FILE_EXTENSION));
    return new ConfigRequest(path.toFile(), filename);
  }

  public void update() {
    // If no provider or if we had an error, do nothing
    if (request.provider == null || broken) return;

    boolean changed = false;
    // For each provider key, see if our config HashMap is missing it
    for (var entry : request.getConfigMap().entrySet()) {
      String key = entry.getKey();
      String value = entry.getValue();
      // If it's missing from 'config', add it
      if (!config.containsKey(key)) {
        config.put(key, value);
        changed = true;
      }
    }

    // If new keys were added, regen the file with createConfig()
    if (changed) {
      try {
        delete();
        createConfig();
      } catch (IOException e) {
        e.printStackTrace();
      }
    }
  }


  private void createConfig() throws IOException {

    // try creating missing files
    request.file.getParentFile().mkdirs();
    Files.createFile(request.file.toPath());

    // write default config data
    PrintWriter writer = new PrintWriter(request.file, StandardCharsets.UTF_8);
    writer.write(request.getConfig());
    writer.close();

  }

  private void loadConfig() throws IOException {
    Scanner reader = new Scanner(request.file);
    for (int line = 1; reader.hasNextLine(); line++) {
      parseConfigEntry(reader.nextLine(), line);
    }
  }

  private void parseConfigEntry(String entry, int line) {
    if (!entry.isEmpty() && !entry.startsWith("#")) {
      String[] parts = entry.split("=", 2);
      if (parts.length == 2) {
        config.put(parts[0].trim(), parts[1].trim());
      } else {
        throw new RuntimeException("Syntax error in config file on line " + line + "!");
      }
    }
  }

  /**
   * Queries a value from config, returns `null` if the
   * key does not exist.
   *
   * @return value corresponding to the given key
   * @see SimpleConfig#getOrDefault
   */
  @Deprecated
  public String get(String key) {
    return config.get(key);
  }

  /**
   * Returns string value from config corresponding to the given
   * key, or the default string if the key is missing.
   *
   * @return value corresponding to the given key, or the default value
   */
  public String getOrDefault(String key, String def) {
    String val = get(key);
    return val == null ? def : val;
  }

  /**
   * Returns integer value from config corresponding to the given
   * key, or the default integer if the key is missing or invalid.
   *
   * @return value corresponding to the given key, or the default value
   */
  public int getOrDefault(String key, int def) {
    try {
      return Integer.parseInt(get(key));
    } catch (Exception e) {
      return def;
    }
  }

  /**
   * Returns boolean value from config corresponding to the given
   * key, or the default boolean if the key is missing.
   *
   * @return value corresponding to the given key, or the default value
   */
  public boolean getOrDefault(String key, boolean def) {
    String val = get(key);
    if (val != null) {
      return val.equalsIgnoreCase("true");
    }

    return def;
  }

  /**
   * Returns double value from config corresponding to the given
   * key, or the default string if the key is missing or invalid.
   *
   * @return value corresponding to the given key, or the default value
   */
  public double getOrDefault(String key, double def) {
    try {
      return Double.parseDouble(get(key));
    } catch (Exception e) {
      return def;
    }
  }

  /**
   * If any error occurred during loading or reading from the config
   * a 'broken' flag is set, indicating that the config's state
   * is undefined and should be discarded using `delete()`
   *
   * @return the 'broken' flag of the configuration
   */
  public boolean isBroken() {
    return broken;
  }

  /**
   * Deletes the config file from the filesystem
   */
  public boolean delete() {
    LOGGER.warn("Config '{}' will rise from the ashes! Please restart game.", request.filename);
    return request.file.delete();
  }

  public interface DefaultConfig {

    static String empty(String namespace) {
      return "";
    }

    String get(String namespace);

    default Map<String, String> getMap() {
      return new HashMap<>();
    }
  }


  public static class ConfigRequest {

    private final File file;
    private final String filename;
    private DefaultConfig provider;

    private ConfigRequest(File file, String filename) {
      this.file = file;
      this.filename = filename;
      this.provider = DefaultConfig::empty;
    }

    /**
     * Sets the default config provider, used to generate the
     * config if it's missing.
     *
     * @param provider default config provider
     * @return current config request object
     * @see DefaultConfig
     */
    public ConfigRequest provider(DefaultConfig provider) {
      this.provider = provider;
      return this;
    }

    /**
     * Loads the config from the filesystem.
     *
     * @return config object
     * @see SimpleConfig
     */
    public SimpleConfig request() {
      return new SimpleConfig(this);
    }

    private String getConfig() {
      return provider.get(filename) + "\n";
    }

    private Map<String, String> getConfigMap() {
      return provider.getMap();
    }

  }

}