package cc.toph.simplycompat.util;

import cc.toph.simplycompat.registry.ItemsRegistry;
import dev.architectury.platform.Platform;

/**
 * The CompatRecord class represents a compatibility record used for managing
 * mod integrations and extending functionality based on certain requirements.
 * This record checks whether a specific mod is present, its version is compatible,
 * and registers tool materials and weapon types if the conditions are met.
 *
 * @param <E>               A type parameter that extends Enum and implements the ToolMaterials
 *                          interface. This defines the supported tool materials for the integration.
 * @param modId             The unique identifier for the mod to be checked.
 * @param requiredVersion   The minimum version of the mod that is required for compatibility.
 * @param toolMaterialClass The class representing the enumeration of tool materials to register.
 */
public record CompatRecord<E extends Enum<E> & ToolMaterials<E>>(
        String modId,
        String requiredVersion,
        Class<E> toolMaterialClass
) {

    /**
     * Checks whether the mod's version is loaded
     * and passes the required version check.
     */
    public boolean passCheck() {
        if (Platform.isModLoaded(modId)) {
            return Platform.getMod(modId)
                           .getVersion()
                           .compareTo(requiredVersion) >= 0;
        }
        return false;
    }

    /**
     * Registers swords for each enum constant in toolMaterialClass,
     * if the material is enabled.
     */
    public void register() {
        if (!passCheck()) {
            return;
        }
        for (WeaponType type : WeaponType.values()) {
            for (E material : toolMaterialClass.getEnumConstants()) {
                if (material.isEnabled()) {
                    ItemsRegistry.registerSword(material, type);
                }
            }
        }
    }
}
